//
// bots, a screensaver by Howard Berkey
//

#include "bots.h"
#include "parseBots.h"

RobotThread *gThr;
BTextControl *gTxt;

class Configurator : public BHandler
{
public:
	static const unsigned long botdebug;
	static const unsigned long botpath;
	virtual void MessageReceived(BMessage *);
};

Configurator *gConfig;
const unsigned long Configurator::botdebug = 0x23232323;
const unsigned long Configurator::botpath = 0x23242526;

float weapon::speed = 10.0;
float weapon::range=512.0;
float Robot::speed = 5.0;
float Robot::scanwidth = 10.0;
float gRate = 1000.0;
char gBotpath[1024];
bool gDebug = false;

instruction gParseTab[] = {
	{"nop", op_nop, 0},
	{"add", op_add, 0},
	{"sub",op_sub, 0},
	{"mul" ,op_mult,0},
	{"div", op_div, 0},
	{"gt", op_gt, 0},
	{"eq", op_eq, 0},
	{"lt", op_lt, 0},
	{"rand", op_rand, 0},
	{"store", op_store, 0},
	{"clr", op_clear, 0},
	{"test",op_test,0},
	{"jsr",op_branch,0},
	{"jmp",op_jump,0},
	{"rts",op_return,0},
	{"move",op_move,1},
	{"scan",op_scan,0},
	{"aim",op_aim,4},
	{"fire",op_fire,9},
	{"mine",op_mine,4},
	{NULL, op_nop,  0}
};


void module_initialize(void *in, long inSize)
{
	strcpy(gBotpath, "/boot/home/botsource");		
}


void module_cleanup(void **out, long *outSize)
{
	*out = NULL;
	*outSize = 0;
}


void module_start_saving(BView	*v)
{	

	v->Window()->Show();

	gThr = new RobotThread;
	gThr->StartSaving(v);	
}


void module_stop_saving()
{
	gThr->StopSaving();
	delete gThr;
	gThr = 0;
}


void module_start_config(BView *v)
{

	//
	// pretty much copied this function from Hiroshi Lockheimer's examples
	//
	if (!v->Window()->Lock())
		return;

	gConfig = new Configurator;
	
	v->Looper()->AddHandler(gConfig);
	
	BStringView *sv = NULL;
	
	sv = new BStringView(BRect(20.0, 20.0, 100.0, 40.0),
								 B_EMPTY_STRING,
								 "bots");
	v->AddChild(sv);
	sv->SetViewColor(v->ViewColor());
	sv->SetFont(be_bold_font);
	
	sv = new BStringView(BRect(30.0, 40.0, 130.0, 52.0),
								 B_EMPTY_STRING,
								 "by Howard Berkey");
	v->AddChild(sv);
	sv->SetViewColor(v->ViewColor());								 
	BMessage *msg = new BMessage(Configurator::botdebug);
	msg->AddInt32("debug", 0);
								 
	BCheckBox *dbg = new BCheckBox(BRect(30.0, 60.0, 140.0, 80.0), "debug",
									"Debug Bots", msg);
	dbg->SetTarget(gConfig);
	v->AddChild(dbg);
	
	msg = new BMessage(Configurator::botpath);
	msg->AddString("botpath", gBotpath);
	
	gTxt = new BTextControl(BRect(20.0, 90.0, 240.0, 120.0), "botpath",
											"Path to Bots", gBotpath, msg);
	gTxt->SetDivider(70.0);
	gTxt->SetTarget(gConfig);
	v->AddChild(gTxt);
		
	v->Window()->Unlock();
}


void  module_stop_config()
{
	BLooper *looper = gConfig->Looper();
	looper->Lock();	
	looper->RemoveHandler(gConfig);
	looper->Unlock();
	delete gConfig;
}


RobotThread::RobotThread()
{

}


					
void RobotThread::StartSaving(BView *view)
{
	m_View = view;
	m_View->Window()->Lock();	
	viewRect = m_View->Bounds();
	arenaRect = viewRect;
	arenaRect.InsetBy(weapon::speed * 1.5, weapon::speed * 1.5);
	m_View->Window()->Unlock();
	loadBots();
	CSaveThread::StartSaving(view);
}


void RobotThread::StopSaving()
{
	CSaveThread::StopSaving();	
	while(m_Robots.size() > 0)
	{
		delete(m_Robots.front());
		m_Robots.pop_front();
	}
	while(m_Weapons.size() > 0)
	{
		m_Weapons.pop_front();
	}
	m_boom = -1;
}


void RobotThread::loadBots()
{
	Robot *r;
	BDirectory dir(gBotpath);
	BEntry ent;
	BPath p;
	
	if(dir.InitCheck() != B_NO_ERROR)
		return;
	while(dir.GetNextEntry(&ent) != ENOENT)
	{
		ent.GetPath(&p);
		
		r = new Robot(this);
		r->parse(p.Path());
		addBot(r);
	}
}


void RobotThread::addBot(Robot *r)
{
	m_Robots.push_back(r);
}


void RobotThread::addShot(weapon &wp)
{
	m_Weapons.push_back(wp);
}


void RobotThread::s_move(BPoint &p, float direction, float speed)
{
	BPoint p1 = p;
	float cd, sd;
	
	direction = (direction / 360.0) * 2 * 3.14159;
	cd = cos(direction);
	sd = sin(direction);
	
	p1.x += (cd * speed);
	p1.y += (sd * speed);
	if(viewRect.Contains(p1))
		p = p1;
}


void RobotThread::s_move(BRect &r, float direction, float speed)
{
	BPoint p = r.LeftTop();
	s_move(p, direction, speed);
	if(arenaRect.Contains(p))
		r.OffsetTo(p);
}


void RobotThread::collision(Robot *a, Robot *b)
{
	weapon w;
	int i;
	if(a == b)
		return;
	m_Robots.remove(a);
	m_Robots.remove(b);
	
	for(i=0;i < 10; i++)
	{
		w.m_shooter = a;
		w.m_dir = rand() % 360 + 1;
		w.m_pos = a->centroid();
		w.m_speed = rand() % 5 + 5;
		w.m_intensity = 255;
		w.m_decay = 10 + (rand() % 10);
		addShot(w);
	}
	for(i=0;i < 10; i++)
	{
		w.m_shooter = b;
		w.m_dir = rand() % 360 + 1;
		w.m_pos = b->centroid();
		w.m_speed = rand() % 5 + 5;
		w.m_intensity = 255;
		w.m_decay = 10 + (rand() % 10);
		addShot(w);
	}

	m_boom = 255;

	delete a;
	delete b;
}


void RobotThread::collision(Robot *r, weapon &s)
{
	weapon w;
	int i;
	
	if(s.m_shooter != r)
	{
		m_Robots.remove(r);
		for(i=0;i < 10; i++)
		{
			w.m_shooter = r;
			w.m_dir = rand() % 360 + 1;
			w.m_speed = rand() % 5 + 5;
			w.m_pos = r->centroid();
			w.m_intensity = 255;
			w.m_decay = 10 + (rand() % 10);
			addShot(w);
		}
		delete r;
		m_boom = 255;
	}

}


double	RobotThread::Save()
{
	botlist::iterator it, tmpit;
	weplist::iterator wt, tmpwt;
	static long victoryCount = 0;
	m_View->Window()->Lock();
	m_View->SetHighColor(254, 0, 254);
	m_View->StrokeRect(arenaRect);	

	if(m_boom >= 0)
	{
		m_View->SetHighColor(m_boom, m_boom, m_boom);
		m_View->FillRect(viewRect);	
		m_boom -= 127;
		if(m_boom < 0)
		{
			m_View->SetHighColor(0, 0, 0);
			m_View->FillRect(viewRect);			
		}
	}
	
	if(victoryCount < 1024)
	{
		//
		// cycle the bots
		//
		for(it = m_Robots.begin(); it != m_Robots.end(); it++)
			{
				(*it)->cpu_cycle(this);
			}
		//
		// move bullets
		//
		for(wt = m_Weapons.begin(); wt != m_Weapons.end(); wt++)
			{
				(*wt).erase(m_View);
				(*wt).m_intensity -= (*wt).m_decay;
				if((*wt).m_speed != 0.0)
					s_move((*wt).m_pos, (*wt).m_dir, (*wt).m_speed);
				if((*wt).m_intensity <= 0 || !arenaRect.Contains((*wt).m_pos))
				{
					tmpwt = wt;
					tmpwt--;
					m_Weapons.erase(wt);
					wt = tmpwt;
				} else
					(*wt).draw(m_View);
			}
		
		// 
		// check collisions 
		//
		for(it = m_Robots.begin(); it != m_Robots.end(); it++)
			{
				botlist::iterator bit;
				weplist::iterator wit;
				bool collided;
				
				collided = false;
				
				for(bit = m_Robots.begin(); bit != m_Robots.end(); bit++)
				{
					if(it == bit)
						continue;
					if((*it)->m_bounds.Intersects((*bit)->m_bounds))
					{
						collision((*it), (*bit));
						it = m_Robots.begin();
						collided = true;
						break;
					}
				}
				
				if(collided == false)
				{
					for(wit = m_Weapons.begin(); wit != m_Weapons.end(); wit++)
					{
						if((*it)->m_bounds.Contains((*wit).m_pos))
						{
							if((*wit).m_shooter == (*it))
								continue;
							collision((*it), (*wit));
							it = m_Robots.begin();
							collided = true;
							break;
						}
					}
				}
				
			}
			
		if(m_Robots.size() == 1)
		{
			m_results = "VICTORY! ";
			m_results += m_Robots.front()->m_name;
			m_results += " wins!";
			victoryCount++;
		} 
		
		if(m_Robots.size() == 0)
		{
			if(victoryCount == 0)
				m_results = "DRAW";		
			victoryCount++;
		} 
	
		if(victoryCount > 0)
		{
			float fx = (float) (rand() % (int) arenaRect.right);
			float fy = (float) (rand() % (int) arenaRect.bottom);

			victoryCount++;					
			m_View->SetHighColor(0, 0, rand() % 255);
			m_View->SetFont(be_bold_font);		
			m_View->SetFontSize(32.0);	
			m_View->DrawString(m_results.c_str(), BPoint(fx, fy));
		}
	} else {
		m_View->SetHighColor(0, 0, 0);
		m_View->FillRect(viewRect);		
		m_Robots.clear();
		m_Weapons.clear();
		loadBots();
		victoryCount = 0;	
	}
	m_View->Window()->Unlock();
	return gRate;
}


void weapon::erase(BView *v)
{
	if(!v->Bounds().Contains(m_pos))
		return;
	v->SetHighColor(0, 0, 0);	
	v->MovePenTo(m_pos);
	v->StrokeLine(m_pos);
}

void weapon::draw(BView *v)
{
	if(!v->Bounds().Contains(m_pos))
		return;
	int intensity = m_intensity;
	
	if(intensity > 255)
		intensity = 255;
	v->SetHighColor(intensity, intensity, intensity);	
	v->MovePenTo(m_pos);
	v->StrokeLine(m_pos);
}


Robot::Robot(RobotThread *r)
{
	float x, y;
	BRect re;
	re = r->arenaRect;
	re.InsetBy(100.0, 100.0);
	x = (float) (rand() % (int) re.right);
	y = (float) (rand() % (int) re.bottom);
	m_red = 20 + rand() % 180;
	m_green = 20 + rand() % 180;
	m_blue = 20 + rand() % 180;
	m_bounds.Set(0.0, 0.0, 20.0, 20.0);
	m_bounds.OffsetBy(BPoint(x, y));
	m_aiming = 0.0;
	m_accumulator = 0.0;
	m_reg1 = 0.0;
	m_reg2 = 0.0;
	m_reg3 = 0.0;
	m_reg4 = 0.0;
	m_pc = m_program.begin();
	m_instr = "";
	
}


void Robot::cpu_cycle(RobotThread *rt)
{
	BPoint tmppt;
//
// erase
//
	rt->m_View->SetHighColor(0, 0, 0);	
	rt->m_View->FillRect(m_bounds);
	rt->m_View->SetFontSize(10.0);	
	rt->m_View->DrawString(m_name.c_str(), m_bounds.LeftTop());
	if(gDebug == true)
	{
		rt->m_View->DrawString(m_instr.c_str());
	
		if(m_pc != m_program.end())
		{
			m_instr = ":"; m_instr += (*m_pc).m_name; m_instr += " ";
			m_instr += (*m_pc).m_arg;	m_instr += " ";
			m_instr += (*m_pc).m_arg1;
		}
	}
	interpret(rt);

//
// draw
//
	BRect oval = m_bounds;
	oval.InsetBy(2.0, 2.0);
	rt->m_View->SetHighColor(m_red, m_green, m_blue);	
	rt->m_View->StrokeEllipse(oval);
	rt->m_View->SetFontSize(10.0);	
	rt->m_View->DrawString(m_name.c_str(), m_bounds.LeftTop());
	if(gDebug == true)
		rt->m_View->DrawString(m_instr.c_str());
	tmppt = centroid();
	rt->m_View->MovePenTo(tmppt);
	rt->s_move(tmppt, m_aiming, 8.0);
	rt->m_View->StrokeLine(tmppt);
	rt->m_View->SetHighColor(254, 254, 254);
	float scarc = -m_scanning;
	scarc -= (Robot::scanwidth / 2.0);
	if(scarc < 0.0) scarc += 360.0;
	rt->m_View->StrokeArc(oval, scarc,Robot::scanwidth);
}


long Robot::longArg(char *arg)
{
	if(strcmp(arg,"r1") == 0)
		return (long) m_reg1;
	if(strcmp(arg,"r2") == 0)
		return (long) m_reg2;
	if(strcmp(arg,"r3") == 0)
		return (long) m_reg3;
	if(strcmp(arg ,"r4") == 0)
		return (long) m_reg4;
	if(strcmp(arg, "acc") == 0)
		return (long) m_accumulator;
		
	return atol(arg);
}

float Robot::floatArg(char *arg)
{
	if(strcmp(arg,"r1") == 0)
		return m_reg1;
	if(strcmp(arg,"r2") == 0)
		return m_reg2;
	if(strcmp(arg,"r3") == 0)
		return m_reg3;
	if(strcmp(arg ,"r4") == 0)
		return m_reg4;
	if(strcmp(arg, "acc") == 0)
		return m_accumulator;

	return (float) longArg(arg);
}

void Robot::storeArg(char *arg, float a)
{
	if(strcmp(arg,"r1") == 0)
		m_reg1 = a;
	if(strcmp(arg,"r2") == 0)
		m_reg2 = a;
	if(strcmp(arg,"r3") == 0)
		m_reg3 = a;
	if(strcmp(arg ,"r4") == 0)
		m_reg4 = a;
	if(strcmp(arg, "acc") == 0)
		m_accumulator = a;

}


void Robot::interpret(RobotThread *rt)
{
	instruction opcode;
	program::iterator it;
	int i;
	
	if(m_pc == m_program.end())
		m_pc = m_program.begin();
	
	if(m_pc == m_program.end())
		return;
		
	opcode = *m_pc;
	m_pc++;
	switch(opcode.m_code)
	{
		case op_nop:
		case op_cyc:
			break;
		case op_add:
			m_accumulator = floatArg(opcode.m_arg) + floatArg(opcode.m_arg1);
			break;
		case op_sub:
			m_accumulator = floatArg(opcode.m_arg) - floatArg(opcode.m_arg1);
			break;
		case op_mult:
			m_accumulator = floatArg(opcode.m_arg) * floatArg(opcode.m_arg1);
			break;		
		case op_div:
			if(floatArg(opcode.m_arg1) != 0.0)
				m_accumulator = floatArg(opcode.m_arg) / floatArg(opcode.m_arg1) ;
			else
				m_accumulator = 0.0;
			break;		
		case op_store:	// store arg x in a register y 
			storeArg(opcode.m_arg, floatArg(opcode.m_arg1));
			break;		
		case op_clear:	// clear register arg
			storeArg(opcode.m_arg, 0.0);
			break;		
		case op_test:	// if accumulator is set, execute next arg else skip it
			if(m_accumulator == 0.0)
			{
				m_pc++;
				while((*m_pc).m_code == op_cyc)
					m_pc++;
			}
			break;		
		case op_branch:	// branch to label
			stackFrame s;
			s.m_pc = m_pc;
			m_stack.push_front(s);
			for(it = m_program.begin(); it != m_program.end(); it++)
			{
				if(strcmp(opcode.m_arg, (*it).m_label) == 0)
				{
					m_pc = it;
					m_pc++;
					break;
				}
			}
			//
			// jump to bad tag resets
			//
			if(m_pc == m_program.end() || it == m_program.end())
			{
				m_pc = m_program.begin();
				m_stack.clear();
			}	
			break;			
		case op_jump:	// jump to label
			for(it = m_program.begin(); it != m_program.end(); it++)
			{
				if(strcmp(opcode.m_arg, (*it).m_label) == 0)
				{
					m_pc = it;
					m_pc++;
					break;
				}
			}
			//
			// jump to bad tag resets
			//
			if(m_pc == m_program.end() || it == m_program.end())
			{
				m_pc = m_program.begin();
				m_stack.clear();
			}	
			break;			
		case op_return: 	// returns from label
			if(m_stack.size() > 0)
			{
				m_pc = m_stack.front().m_pc;
				m_stack.pop_front();
			} else {
				m_pc = m_program.begin();
			}
			break;		
		case op_move:	// arg dir (1-360)
			rt->s_move(m_bounds, floatArg(opcode.m_arg), Robot::speed);
			break;		
		case op_scan:	// arg dir (1-360), stores range in accumulator
			m_scanning = floatArg(opcode.m_arg);
			m_accumulator = p_scan(rt);
			break;		
		case op_aim:		// arg dir (1-360)
			m_aiming = floatArg(opcode.m_arg);
			break;		
		case op_fire:
			weapon w;
			w.m_shooter = this;
			w.m_dir = m_aiming;
			w.m_speed = weapon::speed;
			w.m_pos = centroid();
			w.m_intensity = weapon::range;
			w.m_decay = weapon::speed;
			rt->addShot(w);
			break;		
		case op_mine:
			weapon m;
			m.m_shooter = this;
			m.m_dir = 0.0;
			m.m_speed = 0.0;
			m.m_intensity = 512;
			m.m_decay = 1;
			m.m_pos = centroid();
			rt->addShot(m);
			break;	
		case op_rand:
			m_accumulator = (float) (rand() % longArg(opcode.m_arg));
			break;	
		case op_gt:
			if(floatArg(opcode.m_arg) > floatArg(opcode.m_arg1))
				m_accumulator = 1.0;
			else
				m_accumulator = 0.0;
			break;
		case op_lt:
			if(floatArg(opcode.m_arg) < floatArg(opcode.m_arg1))
				m_accumulator = 1.0;
			else
				m_accumulator = 0.0;
			break;
		case op_eq:
			if(floatArg(opcode.m_arg) == floatArg(opcode.m_arg1))
				m_accumulator = 1.0;
			else
				m_accumulator = 0.0;
			break;
		}
	
}

float Robot::p_scan(RobotThread *rt)
{
	float bearing = 0.0, dist;
	BPoint p1, p2;
	RobotThread::botlist::iterator bit;
	
	p1 = centroid();
	
	for(bit = rt->m_Robots.begin(); bit != rt->m_Robots.end(); bit++)
	{
		if(this == *bit)
			continue;
		p2 = (*bit)->centroid();
		p2.x -= p1.x;
		if(p2.x == 0.0)
			p2.x = 0.1;
		p2.y -= p1.y;
		dist = (p2.x * p2.x) + (p2.y * p2.y);
		dist = sqrt(dist);
		bearing = acos(p2.x/dist);
		bearing /= (2.0 * 3.14159);
		bearing *= 360.0;
		if(p2.y < 0.0)
			bearing = 360.0 - bearing;
		
		if((bearing > (m_scanning - (Robot::scanwidth / 2.0))) && (bearing < (m_scanning + (Robot::scanwidth / 2.0))))
		{
			return dist;
		}
	}
	
	return 0.0;
	
}

instruction PARSE(char *n)
{
	int i;
	for(i=0;gParseTab[i].m_name != NULL; i++)
	{
		if(gParseTab[i].m_name != NULL)
			if(strcmp(n, gParseTab[i].m_name) == 0)
				return gParseTab[i];
	}
	//
	// if it matches no instructions, it must be a label
	//
	instruction in = gParseTab[0];
	strcpy(in.m_label, n);
	return in; // NOP
}


void Robot::parse(const char *filepath)
{
	FILE *f;
	char buf[255], *tok;
	instruction ins;
	float x, y;
	instruction ci;

	ci.m_code = op_cyc;
	ci.m_arg1[0] = 0;
	ci.m_arg[0] = 0;	
	
	ins.m_label[0] = 0;
	f = fopen(filepath, "r");
	if(f == NULL)
		return;
		
	fgets(buf, sizeof(buf), f);
	tok = strtok(buf, " \n");
	if(tok == NULL)
		return;	
	m_name = tok;
		
	while(fgets(buf, sizeof(buf), f) != NULL)
	{
		ins.m_arg[0] = 0;
		ins.m_arg1[0] = 0;
		tok = strtok(buf, " \n");
		if(tok == NULL)
			continue;
		ins = PARSE(tok);
		tok = strtok(NULL, " \n");
		if(tok != NULL)
		{
			strcpy(ins.m_arg, tok);
			tok = strtok(NULL, " \n");		
			if(tok != NULL)
				strcpy(ins.m_arg1, tok);
		}
		m_program.push_back(ins);	
		for(int i=0;i<ins.m_cycles; i++)
		{
			ci.m_name = ins.m_name;
			m_program.push_back(ci);
		}	
	}
	m_program.push_back(PARSE("nop"));
	fclose(f);
}

void Configurator::MessageReceived(BMessage *msg)
{
	switch (msg->what){
		case Configurator::botpath:
			strcpy(gBotpath, gTxt->Text());
			break;
		case Configurator::botdebug:
			if(gDebug == false)
			{
				gRate = 1000000.0;
				gDebug = true;
			}
			else
			{
				gRate = 1000.0;
				gDebug = false;
			}

			break;
	}
}
